/*****************************************************************************
 * main.c
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 *
 * Copyright (C) 2000-2002 Chris Pinkham
 * Released under the terms of the GPL.
 * *NO WARRANTY*
 *
 * cpinkham@infi.net, cpinkham@bc2va.org
 * http://www4.infi.net/~cpinkham/gyach/
 *****************************************************************************/

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <unistd.h>
#include <time.h>
#include <gtk/gtk.h>

#ifdef USE_PTHREAD_CREATE
#include <pthread.h>
#endif

#include "gyach.h"

#ifndef OS_WINDOWS
#	include <X11/Xlib.h>
#	include <gdk/gdkx.h>
#endif

#include <errno.h>

#include "aliases.h"
#include "callbacks.h"
#include "commands.h"
#include "friends.h"
#include "gyach_int.h"
#include "ignore.h"
#include "images.h"
#include "interface.h"
#include "main.h"
#include "support.h"
#include "users.h"
#include "util.h"
#include "yahoochat.h"

/* ------------------------------------------------------------------------ */ 
 
typedef enum
{ Medium, Bold }
FontWeight;
typedef enum
{ Italic, Roman }
FontSlant;

static GdkFont *get_font(const char *family, const short size,
	const FontWeight weight, const FontSlant slant)
{
	GdkFont *font = NULL;
	char XLFD[1000];
	char my_font_name[129];
	char *ptr;
	char *from;
	char *to;

#ifndef OS_WINDOWS
	XFontStruct *fontOK = (XFontStruct *) NULL;
#endif

	DBG( 21, "get_font( '%s', %d, %d, %d )\n", family, size, weight, slant );

	if ( font ) {
		gdk_font_unref( font );
		font = NULL;
	}

	strcpy( my_font_name, font_name );
	strcpy(XLFD, "-*");			/* foundry */

	if (family && *family)
	{
		char *p = NULL, *lowerFace = NULL;

		strcat(XLFD, "-");
		p = lowerFace = strdup(family);
		while (p && *p)
		{
			int c = tolower(*p);

			*p = c;
			++p;
		}
		strcat(XLFD, lowerFace);
		free(lowerFace);
	}
	else
	{
		strcat(XLFD, "-times");	/* default family, I should really get this from the current theme */
	}

	switch (weight)
	{
		case Bold:
			strcat(XLFD, "-bold");	/* weight */
			if (( ptr = strstr( my_font_name, "medium" ))) {
				strcpy( ptr, "bold" );
				to = ptr + 4;
				from = ptr + 6;
				while( *from ) {
					*(to++) = *(from++);
				}
				*to = '\0';
			}
			break;

		case Medium:
		default:
			strcat(XLFD, "-medium");	/* weight */
			if (( ptr = strstr( my_font_name, "bold" ))) {
				from = strstr( font_name, "bold" ) + 4;
				strcpy( ptr, "medium" );
				to = ptr + 6;
				while( *from ) {
					*(to++) = *(from++);
				}
				*to = '\0';
			}
			break;
	}

	switch (slant)
	{
		case Italic:
			strcat(XLFD, "-i");	/* slant : italic */
			break;

		case Roman:
		default:
			strcat(XLFD, "-r");	/* slant : roman */
			break;
	}

	strcat(XLFD, "-normal");	/* set width */
	strcat(XLFD, "-*");			/* additional style (nil) */

	/* dunno why, but he appears to have had these next 2 fields reversed, it */
	/* works better this way. */
	/*
	if (size > -1)
	{
		char sizeS[10];

		sprintf(sizeS, "-%d", 10 * size);
		strcat(XLFD, sizeS);
	}
	else
	{
		strcat(XLFD, "-*");
	}
	*/

	{
		char sizeS[10];

		sprintf(sizeS, "-%d", font_size);
		strcat(XLFD, sizeS);
	}
	strcat(XLFD, "-*");			/* pixel size (don't care) */

	strcat(XLFD, "-*");			/* resolution X */
	strcat(XLFD, "-*");			/* resolution Y */
	strcat(XLFD, "-*");			/* proportional spacing */
	strcat(XLFD, "-*");			/* Average width */
	strcat(XLFD, "-*");			/* Charset registry */
	strcat(XLFD, "-*");			/* Charset encoding */

	/*
	printf( "XLFD: %s\n", XLFD );
	printf( "  my: %s\n", my_font_name );
	fflush( stdout );
	*/

	if ( ! show_fonts ) {
		strcpy( XLFD, my_font_name );
	}

	/* Test load the font.  If it fails, fontOK is null, and get on with
	 * life after giving the user an error.
	 * If it is successful, immediately unload the font, since we're using
	 * GTK+ here, mmmk?
	 */
#ifndef OS_WINDOWS
	fontOK = XLoadQueryFont(GDK_DISPLAY(), XLFD);

	if (fontOK == (XFontStruct *) NULL || (!fontOK))
	{
		/* Font was no good - you can't load this font. */
		font = NULL;
	}
	else
	{
		/* Remember to free this since XLoadQueryFont returns allocated memory */
		XFreeFont(GDK_DISPLAY(), fontOK);
		font = gdk_font_load(XLFD);
	}
#endif

	if ( ! font ) {
		strcpy( XLFD, "-*-times-medium-r-normal-*-14-*-*-*-*-*-*-*" );
		font = gdk_font_load( XLFD );
	}

	return font;
}

static GdkFont *getFont(int bold, int italic, int underline,
	const char *family, const char *sizeS)
{
	static GdkFont *font = NULL;
	FontWeight weight;
	FontSlant slant;
	short size = 0;

	DBG( 21, "getFont( %d, %d, %d, '%s', '%s' )\n",
		bold, italic, underline, family, sizeS );

	if (sizeS && *sizeS)
	{
		size = atoi(sizeS);
	}
	weight = bold ? Bold : Medium;
	slant = italic ? Italic : Roman;

	if (NULL != font)
	{
		gdk_font_unref(font);
		font = NULL;
	}

	if (NULL == (font = get_font(family, size, weight, slant)))
	{
		/* If the getting the actual font fails,
		   first try a few sizes smaller and
		   then try a blank family */
		short trySize = size - 1, endTrySize = (size - 4);

		if (family && *family)
		{
			while ((trySize > 0) &&
				(NULL == (font = get_font(family, trySize, weight, slant))) &&
				(trySize > endTrySize))
			{
				--trySize;
			}
		}

		if (NULL == font)
		{
			trySize = size;
			while ((trySize > 0) &&
				(NULL == (font = get_font(NULL, trySize, weight, slant))) &&
				(trySize > endTrySize))
			{
				--trySize;
			}
		}

	}

	if (NULL == font)
	{
		gdk_font_ref(font);
	}

	return (font);
}

static int bold = FALSE, italic = FALSE, underline = FALSE, url = FALSE;

void append_to_textbox(GtkWidget *window, GtkWidget *textbox, char *text)
{
	int i = 0;
	GdkColor colorspace;
	GdkColor *color = &colorspace;
	GdkFont *font = NULL;
	char face[1000] = "", size[1000] = "";
	int tmp_color;
	char *tmp_ptr;
	char buf[1024];
	int scroll = FALSE;
	unsigned int r, g, b;
	unsigned int or, og, ob;
	char *sound = NULL;
	struct stat sbuf;
#ifdef USE_PTHREAD_CREATE
	pthread_t prof_thread;
#endif


	DBG( 21, "append_to_textbox( %p, %p, '%s' )\n",
		window, textbox, text );

	*face = *size = '\0';

	strcpy( face, font_family );

	font = getFont(bold, italic, underline, face, size);

	if ( force_lowercase )
		lower_str( text );

	strip_junk( text );
	convert_tags( text );

	if ( recv_sound ) {
		sound = recv_sound;
		recv_sound = NULL;

#ifdef OS_WINDOWS
		sprintf( buf, "./gyach/sounds/%s.wav", sound );
#else
		sprintf( buf, "%s/.gyach/sounds/%s.wav", getenv( "HOME" ), sound );
#endif

		free( sound );
		sound = strdup( buf );

		if ( ! stat( buf, &sbuf )) {
#ifdef USE_PTHREAD_CREATE
			pthread_create( &prof_thread, NULL, play_sound, (void *)sound );
#else
			play_sound( (void *)sound );
#endif
		}
	}

	if (( ! textbox ) &&
		( ct_scroll_at_bottom())) {
		scroll = TRUE;
	}

	/* Reset all attributes - currently just color */
	r = g = b = 0;
	or = og = ob = 0;

	/* freeze before printing to speed up display, especially on blended */
	/* colors or for people who switch colors every letter */
	if ( ! textbox )
		ct_freeze();

	i = 0;
	while (text && i < strlen(text))
	{
		if (!strncmp(&text[i], YAHOO_FONT_TAG_START,
				strlen(YAHOO_FONT_TAG_START)))
		{
			int closeFound = FALSE;

			i += strlen(YAHOO_FONT_TAG_START);
			while (!closeFound)
			{
				if (!strncmp(&text[i], YAHOO_FONT_FACE_START,
						strlen(YAHOO_FONT_FACE_START)))
				{
					int k = 0;

					i += strlen(YAHOO_FONT_FACE_START);
					while (text && i < strlen(text) && text[i] != '"')
					{
						face[k] = text[i];
						++i;
						++k;
						face[k] = '\0';
					}
					while (text && i < strlen(text)
						&& NULL != strchr("\" \t", text[i]))
					{
						++i;
					}
				}

				if (text && '>' == text[i])
				{
					++i;
					closeFound = TRUE;
					continue;
				}

				if (!strncmp(&text[i], YAHOO_FONT_SIZE_START,
						strlen(YAHOO_FONT_SIZE_START)))
				{
					int k = 0;

					i += strlen(YAHOO_FONT_SIZE_START);
					while (text && i < strlen(text) && text[i] != '"')
					{
						size[k] = text[i];
						++i;
						++k;
						size[k] = '\0';
					}
					while (text && i < strlen(text)
						&& NULL != strchr("\" \t", text[i]))
					{
						++i;
					}
				}
			}

			if ( show_fonts ) /* should we use their font? */
			{
				font = getFont(bold, italic, underline, face, size);
			}
		}
		else if (( text[i] == 033 ) && ( text[i+1] == '[' ))
		{
			i += 2;  /* skip over the escape and [ */
			if ( text[i+1] == ';' ) {  /* we have a bold/not bold */
				if ( text[i] == '0' ) {
					bold = FALSE;
				} else {
					bold = TRUE;
				}
			    font = getFont(bold, italic, underline, face, size);
				i += 2;
			}
			if ( text[i] == 'x' ) {
				/* turn something OFF */
				i++;
				switch ( text[i] ) {
					case '1': bold = FALSE ;
							  font = getFont(bold, italic, underline, face,
							  		size);
					          break;
					case '2': italic = FALSE;
							  font = getFont(bold, italic, underline, face,
							  		size);
					          break;
					case '4': underline = FALSE;
							  font = getFont(bold, italic, underline, face,
							  		size);
					          break;
					case 'l': url = FALSE ;
							  r = or; g = og; b = ob;
							  font = getFont(bold, italic, underline, face,
							  	size);
							  break;
				}
				i += 2;
				continue;
			} else if ( text[i] == '#' ) {
				/* we found a blended color */
				char tmp_int[3];

				if ( show_blended_colors ) {
					or = r; og = g; ob = b;

					tmp_int[2] = '\0';

					tmp_int[0] = text[ i + 1 ];
					tmp_int[1] = text[ i + 2 ];
					sscanf( tmp_int, "%x", &r );

					tmp_int[0] = text[ i + 3 ];
					tmp_int[1] = text[ i + 4 ];
					sscanf( tmp_int, "%x", &g );

					tmp_int[0] = text[ i + 5 ];
					tmp_int[1] = text[ i + 6 ];
					sscanf( tmp_int, "%x", &b );
				}

			    i += 8;
			    continue;
			} else {
				tmp_color = atoi( &text[i] );
				switch ( tmp_color ) {
					case 0: if (( text[i] == 'l' ) && ( text[i+1] == 'm' )) {
							  if ( show_colors ) {
					        	url = TRUE ;
								or = r; og = g; ob = b;
								r = 0; g = 0; b = 0xff;
							  }
							  i += 2;
							  continue;
							} else if ( text[i+1] == 'm' ) {
							  italic = FALSE;
							  bold = FALSE;
							  underline = FALSE;
							  url = FALSE;
							  font = getFont(bold, italic, underline, face,
							  	size);

							  /* black */
							  or = r; og = g; ob = b;
							  r = 0; g = 0; b = 0;
							  i += 2;
							  continue;
							}
						    break;
					case 1: bold = TRUE ;
							if ( show_fonts )
								font = getFont(bold, italic, underline, face,
									size);
							i += 2;
							continue;
					        break;
					case 2: italic = TRUE;
							if ( show_fonts )
								font = getFont(bold, italic, underline, face,
									size);
							i += 2;
							continue;
					        break;
					case 4: underline = TRUE;
							if ( show_fonts )
								font = getFont(bold, italic, underline, face,
									size);
							i += 2;
							continue;
					        break;
					case 30: /* black */
							if ( show_colors ) {
								or = r; og = g; ob = b;
								r = 0; g = 0; b = 0;
							}
							i += 3;
							break;
					case 31: /* blue */
							if ( show_colors ) {
								or = r; og = g; ob = b;
								r = 0; g = 0; b = 0xff;
							}
							i += 3;
						    break;
					case 32: /* cyan */
							if ( show_colors ) {
								or = r; og = g; ob = b;
								r = 0; g = 0xb2; b = 0xb2;
							}
							i += 3;
						    break;
					case 33: /* gray */
							if ( show_colors ) {
								or = r; og = g; ob = b;
								r = 0x80; g = 0x80; b = 0x80;
							}
							i += 3;
						    break;
					case 34: /* green */
							if ( show_colors ) {
								or = r; og = g; ob = b;
								r = 0; g = 0xc2 ; b = 0;
							}
							i += 3;
						    break;
					case 35: /* pink */
							if ( show_colors ) {
								or = r; og = g; ob = b;
								r = 0xff; g = 0xaf; b = 0xaf;
							}
							i += 3;
						    break;
					case 36: /* purple */
							if ( show_colors ) {
								or = r; og = g; ob = b;
								r = 0xb2; g = 0; b = 0xb2;
							}
							i += 3;
						    break;
					case 37: /* yellow */
							if ( show_colors ) {
								or = r; og = g; ob = b;
								r = 0xff; g = 0xff ; b = 0;
							}
							i += 3;
						    break;
					case 38: /* red */
							if ( show_colors ) {
								or = r; og = g; ob = b;
								r = 0xff ; g = 0; b = 0;
							}
							i += 3;
						    break;
					case 39: /* olive */
							if ( show_colors ) {
								or = r; og = g; ob = b;
								r = 0x54; g = 0x6b ; b = 0x50;
							}
							i += 3;
						    break;
				}
			}
			continue;
		} else if ( text[i] == 033 ) {
			i++;
		}

		if (i >= strlen(text))
		{
			break;
		}

		if ( text[i] == 033 ) {
			continue;
		}

		/* print out as much as we can */
		if (( tmp_ptr = strchr( &text[i], 033 )) != NULL ) {
			tmp_color = tmp_ptr - &text[i];
		} else {
			tmp_color = strlen( &text[i] );
		}

		if ( textbox ) {
			color->red   = r * ( 65535 / 255 );
			color->green = g * ( 65535 / 255 );
			color->blue  = b * ( 65535 / 255 );
			gdk_colormap_alloc_color(
				gtk_widget_get_colormap(GTK_WIDGET(textbox)),
				color, FALSE, TRUE);
			gtk_text_insert(GTK_TEXT(textbox), font, color,
				NULL, (gchar *) & text[i], tmp_color );
		} else {
			ct_set_color( r, g, b );
			ct_set_font( font );

			tmp_ptr = strstr( text, ":\n" );
			/* attempt to display /exec-ed stuff in fixed width font */
			if (( tmp_ptr != NULL ) &&
				( tmp_ptr != ( text + strlen( text ) - 2 ))) {
				ct_append_fixed( &text[i], tmp_color );
			} else {
				ct_append( &text[i], tmp_color );
			}
		}

		if ( capture_fp ) {
			fwrite( &text[i], 1, tmp_color, capture_fp );
		}

		i += tmp_color;
	}

	if ( ! textbox )
		ct_thaw();

	if( scroll )
		ct_scroll_to_bottom();

	if ( capture_fp )
		fflush( capture_fp );

	DBG( 22, "RETURN append_to_textbox()\n" );
}

void append_to_textbox_color(GtkWidget *win, GtkWidget *tb, char *text) {
	int old_colors = show_colors;
	int old_fonts = show_fonts;
	int old_lower = force_lowercase;

	DBG( 21, "append_to_textbox_color( %p, %p, '%s' )\n", win, tb, text );

	show_colors = 1;
	show_fonts = 1;
	force_lowercase = 0;

	/* reset things to normal */
	if ( bold )
		append_to_textbox( win, tb, YAHOO_STYLE_BOLDOFF );

	if ( italic )
		append_to_textbox( win, tb, YAHOO_STYLE_ITALICOFF );

	if ( underline )
		append_to_textbox( win, tb, YAHOO_STYLE_UNDERLINEOFF );

	if ( url )
		append_to_textbox( win, tb, YAHOO_STYLE_URLOFF );

	/* now send our text */
	append_to_textbox( win, tb, text );

	show_colors = old_colors;
	show_fonts = old_fonts;
	force_lowercase = old_lower;

	DBG( 22, "RETURN append_to_textbox_color()\n" );
}

/* ------------------------------------------------------------------------ */ 
/* send a chat command 
 * Note: it modifies the string in place 
 * return FALSE if we should quit 
 */ 
int chat_command( char *cmd ) 
{ 
	int r    = TRUE; 
	char *txt = ""; 
	char *args;
	char buf[1025] = "";
	regex_t smiley_regex;

	DBG( 21, "chat_command( '%s' )\n", cmd );

	if ( prof_text )
		display_profile();

	cmd = skip_whitespace( cmd );

	auto_away_time = time(NULL) + ( auto_away * 60 );

	switch(cmd[0]) { 
		case '\0': 
			return( TRUE );
			break; 
		case ':':
			/* differentiate between smileys and emotes */
			regcomp( &smiley_regex, "^:[)({}|<>DP] ", REG_NOSUB );
			
			if ( ! regexec( &smiley_regex, cmd, 0, NULL, 0 )) {
				/* it's a smiley */
				regfree( &smiley_regex );
				txt = cmd;
				subst_escs( txt );
			} else {
				/* it's not */
				regfree( &smiley_regex );
				sprintf( buf, "/emote %s", cmd+1 );
				try_command( buf );
				return( TRUE );
			}
			break;
		case '!':
			if ( cmd[1] == '!' ) {
				sprintf( buf, "/execlocal %s", cmd+2 );
			} else {
				sprintf( buf, "/exec %s", cmd+1 );
			}
			try_command( buf );
			return( TRUE );
			break;

		case '/': 
			if ( try_command( cmd ))
				return( TRUE );

			args = find_whitespace( cmd + 1);
			if ( *args ) {
				*args = '\0';
				args++;
				args = skip_whitespace( args );
			}

			if ( check_alias( cmd + 1, args ))
				return( TRUE );

			sprintf( buf, "*** Gyach: unknown command '%s'\n", cmd );
			append_to_textbox( chat_window, NULL, buf );
			return( FALSE );
        	break; 
 
		default: 
			if (( cmd[0] == '\\' ) &&
				(( cmd[1] == '/' ) ||
				 ( cmd[1] == ':' ) ||
				 ( cmd[1] == '!' ))) {
				txt = cmd +1;
			} else {
				txt = cmd; 
			}

			subst_escs( txt );
			break;
    } 

	if ( use_color ) {
		if ( ! strcmp( use_color, "black" )) {
			strcat( buf, YAHOO_COLOR_BLACK );
		} else if ( ! strcmp( use_color, "blue" )) {
			strcat( buf, YAHOO_COLOR_BLUE );
		} else if ( ! strcmp( use_color, "cyan" )) {
			strcat( buf, YAHOO_COLOR_CYAN );
		} else if ( ! strcmp( use_color, "gray" )) {
			strcat( buf, YAHOO_COLOR_GRAY );
		} else if ( ! strcmp( use_color, "green" )) {
			strcat( buf, YAHOO_COLOR_GREEN );
		} else if ( ! strcmp( use_color, "pink" )) {
			strcat( buf, YAHOO_COLOR_PINK );
		} else if ( ! strcmp( use_color, "purple" )) {
			strcat( buf, YAHOO_COLOR_PURPLE );
		} else if ( ! strcmp( use_color, "yellow" )) {
			strcat( buf, "\033[#ffff00m" );
		} else if ( ! strcmp( use_color, "orange" )) {
			strcat( buf, "\033[#ffc800m" );
		} else if ( ! strcmp( use_color, "red" )) {
			strcat( buf, YAHOO_COLOR_RED );
		} else if ( ! strcmp( use_color, "custom" )) {
			sprintf( buf, "\033[#%2x%2x%2xm",
				custom_color_red, custom_color_green, custom_color_blue );
		}
	}

	if ( use_bold )
		strcat( buf, YAHOO_STYLE_BOLDON );

	if ( use_italics )
		strcat( buf, YAHOO_STYLE_ITALICON );

	if ( use_underline )
		strcat( buf, YAHOO_STYLE_UNDERLINEON );

	if ( filter_command ) {
		strcat( buf, filter_text( txt ));
	} else {
		strcat( buf, txt );
	}

    if ( *buf ) {
		if( send_avatar ) {
			strcat( buf, AVATAR_START );
			strcat( buf, send_avatar );
			strcat( buf, AVATAR_SEND_END );
#ifndef OS_WINDOWS
			if ( show_avatars ) {
				display_avatar( ymsg_sess->user, send_avatar );
			}
#endif
		}

		ymsg9_comment( ymsg_sess, buf );
		r = 1;
	}
 
	DBG( 22, "RETURN chat_command() == %d\n", r );

    return r; 
} 
 
char ymsg_fields[300][1024];

int split( char *str, char *connector )
{
	char	*last;
	char	*orig;
	char	*ptr;
	char	tmp_str[ 2048 ];
	int	result = 0;
	int i = 0;

	strcpy( tmp_str, str );
	str = tmp_str;

	last = orig = str;

	while (( ptr = strstr( str, connector ))) 
	{
		*ptr = '\0';
		strcpy( ymsg_fields[result], last );
		str = last = ptr + strlen( connector );
		result++;
	}

	if ( last != orig ) {
		strcpy( ymsg_fields[result], last );
	} else {
		result--;
	}

	strcpy( ymsg_fields[result], "__END__" );
	strcpy( ymsg_fields[result+1], "__END__" );

	i = 0;
	while( strcmp( ymsg_fields[i], "__END__" )) {
		if (( debug_packets ) &&
			( strcmp( ymsg_fields[i], "117" ))) {
			fprintf( stderr, "%-3s = '%s'\n",
				ymsg_fields[i], ymsg_fields[i+1] );
			fflush( stderr );
		}
		i += 2;
	}

	return( result + 1 );
}


char *ymsg9_field_p( char *key ) {
	int i = 0;
	static char result[2048];

	result[0] = '\0';
	while( strcmp( ymsg_fields[i], "__END__" )) {
		if ( ! strcmp( ymsg_fields[i], key )) {
			if ( result[0] ) {
				strcat( result, "|" );
			}
			strcat( result, ymsg_fields[i+1] );
		}
		i += 2;
	}
	return( result );
}

char *ymsg9_field( char *key ) {
	int i = 0;
	static char result[2048];

	result[0] = '\0';
	while( strcmp( ymsg_fields[i], "__END__" )) {
		if ( ! strcmp( ymsg_fields[i], key )) {
			if ( result[0] ) {
				strcat( result, "," );
			}
			strcat( result, ymsg_fields[i+1] );
		}
		i += 2;
	}
	return( result );
}


void show_yahoo_packet() {
	static char buf[1024];  /* these are static so they aren't on the stack */
	static char tmp[2048];
	static char tmp2[2048];
	static char tmp3[2048];
	char *lf;
	char *ptr;
	char *src;
	char *end;
	int lf_count = 0;
	int last;
	time_t time_now;
	int hours;
	int minutes;
	int seconds;
	struct tm *tm_now;
	GList *pm_lptr;
	PM_SESSION *pm_sess;
	int display;
	int accept_pm;
	int count;
	time_t exp_time;

	if ( ymsg_sess->pkt.size ) {
		split( ymsg_sess->pkt.data, "\xC0\x80" );
	} else {
		strcpy( ymsg_fields[0], "__END__" );
		strcpy( ymsg_fields[1], "__END__" );
	}

	switch( ymsg_sess->pkt.type ) {
		case YMSG9_GET_KEY:
				connect_time = time(NULL);

				strcpy( tmp, ymsg9_field( "94" ));
				ymsg9_login( ymsg_sess, tmp );

				gtk_statusbar_pop( GTK_STATUSBAR(chat_status), st_cid );
				gtk_statusbar_push( GTK_STATUSBAR(chat_status), st_cid,
					"Logging in to chat server..." );

				sprintf(buf ,"Chat started, type /help for help\n"); 
				append_to_textbox( chat_window, NULL, buf );

				/* reset cookies */
				ymsg_sess->cookie[0] = '\0';
				break;

		case YMSG9_COOKIE:
				/* set online if we got the right cookie packet */
				if ( strcmp( ymsg9_field( "3" ), "" )) {
					gtk_statusbar_pop( GTK_STATUSBAR(chat_status), st_cid );
					gtk_statusbar_push( GTK_STATUSBAR(chat_status), st_cid,
						"Setting status to Online..." );

					ymsg9_online( ymsg_sess );
				}

				/* store the cookies */
				if ( ymsg_sess->cookie[0] ) {
					strcat( ymsg_sess->cookie, ";" );
					strcat( ymsg_sess->cookie, ymsg9_field_p( "59" ));
				} else {
					strcpy( ymsg_sess->cookie, ymsg9_field_p( "59" ));
				}
				ptr = ymsg_sess->cookie;
				ptr = strchr( ptr, '	' ); /* a tab */
				while( ptr ) {
					*ptr = '=';
					ptr = strchr( ptr, '|' );

					if ( ptr ) {
						*ptr = ';';
						ptr = strchr( ptr, '	' ); /* a tab */
					}
				}

				/* set our buddy list, 88 might be ignore list */
				if ( strcmp( ymsg9_field( "87" ), "" )) {
					populate_friend_list( ymsg9_field( "87" ));
				}
				/* merge yahoo ignore list into local list */
				if ( strcmp( ymsg9_field( "88" ), "" )) {
					strcpy( tmp, ymsg9_field( "88" ));
					ptr = tmp;
					while( ptr ) {
						end = strchr( ptr, ',' );
						if ( end )
							*end = '\0';

						if ( ! ignore_check( ptr ))
							ignore_toggle( ptr );

						if ( end ) {
							ptr = end + 1;
						} else {
							ptr = NULL;
						}
					}
				}
				break;

		case YMSG9_ONLINE:
				ymsg9_join( ymsg_sess );

				gtk_statusbar_pop( GTK_STATUSBAR(chat_status), st_cid );
				sprintf( buf, "Current Room: %s", ymsg_sess->room );
				gtk_statusbar_push( GTK_STATUSBAR(chat_status),
					st_cid, buf );
				break;

		case YMSG9_MAIL:
				sprintf( buf,
					"Gyach - You have %s%s%s email(s).  "
						"http://mail.yahoo.com\n",
					YAHOO_COLOR_RED, ymsg9_field( "9" ), YAHOO_COLOR_BLACK );
				append_to_textbox( chat_window, NULL, buf );
				break;

		case YMSG9_COMMENT: 
				/* user comment/emote/thought */
				strcpy( tmp, ymsg9_field( "109" ));
				strcpy( tmp2, ymsg9_field( "117" ));

				/* set auto-away if needed */
				/* should put this in the event loop somehow */
				if (( ! my_status ) &&
					( auto_away ) &&
					( time(NULL) > auto_away_time )) {
					cmd_away( "11" );
				}

				if ( ignore_check( tmp )) {
					return;
				}

				if ( mute_check( tmp )) {
					return;
				}

				ptr = get_last_comment( tmp );
				if (( first_post_is_url ) &&
					(( strstr( tmp2, "http://" )) ||
					 ( strstr( tmp2, "ftp://" )) ||
					 ( strstr( tmp2, " www." ))) &&
					(( ptr == NULL ) ||
					 ( !strcmp( ptr, "Entered Room" )))) {
					if ( ! ignore_check( tmp )) {
						ignore_toggle( tmp );
					}
					sprintf( buf, "Gyach - auto-ignored %s%s%s, first "
						"post was a URL: '%s'\n",
						YAHOO_COLOR_RED, tmp, YAHOO_COLOR_BLACK, tmp2 );
					append_to_textbox( chat_window, NULL, buf );
					return;
				}

				if (( ptr = auto_ignore_check( tmp2 )) != NULL ) {
					if ( ! ignore_check( tmp )) {
						ignore_toggle( tmp );
					} else {
						return;
					}

					sprintf( buf, "Gyach - auto-ignored %s%s%s, regex '%s'\n",
						YAHOO_COLOR_RED, tmp, YAHOO_COLOR_BLACK, ptr );
					append_to_textbox( chat_window, NULL, buf );

					if ( ! disp_auto_ignored ) {
						return;
					}
				} 

				/* keep track of last thing the user said */
				count = set_last_comment( tmp, tmp2 );
				if (( ignore_on_mults ) &&
					( count > ignore_on_mults ) &&
					( ! ignore_check( tmp ))) {
					ignore_toggle( tmp );
					sprintf( buf,
						"Gyach - auto-ignored %s%s%s, same post "
						"%d times: \"%s\"\n", YAHOO_COLOR_RED, tmp,
						YAHOO_COLOR_BLACK, count, tmp2 );
					append_to_textbox( chat_window, NULL, buf );
				}
				if (( mute_on_mults ) &&
					( count > mute_on_mults ) &&
					( ! mute_check( tmp ))) {
					mute_toggle( tmp );
					sprintf( buf,
							"Gyach - auto-muted %s%s%s, same post "
							"%d times: \"%s\"\n", YAHOO_COLOR_RED, tmp,
							YAHOO_COLOR_BLACK, count, tmp2 );
							append_to_textbox( chat_window, NULL, buf );
				}
				if (( suppress_mult ) &&
					( count > 1 )) {
					return;
				}

				/* check for lots of linefeeds */
				if ( limit_lfs ) {
					lf = strchr( tmp2, '\n' );
					while( lf ) {
						lf_count++;
						lf = strchr( lf + 1, '\n' );
					}
					if ( lf_count > limit_lfs ) {
						lf = strchr( tmp2, '\n' );
						while( lf ) {
							*lf = '|';
							lf = strchr( lf + 1, '\n' );
						}
					}
				}

				/* need to make fetch_avatar windows-safe */
#ifndef OS_WINDOWS
				if ( show_avatars ) {
					/* check for an avatar and update if needed */
					if (( ptr = strstr( tmp2, AVATAR_START )) != NULL ) {
						/* avatar exists so try to display if unset */
						ptr += strlen( AVATAR_START );
						end = strstr( ptr, AVATAR_END );
						if ( ! end ) {
							end = strchr( ptr, '>' );
						}
						if ( end ) {
							*end = '\0';
							display_avatar( tmp, ptr );
						}
					}
				}
#endif

				strcpy( tmp2, ymsg9_field( "117" ));
				if ( chat_timestamp ) {
					time_now = time( NULL );
					tm_now = localtime( &time_now );
					sprintf( buf, "%s[%d/%d/%d %02d:%02d] ",
						YAHOO_COLOR_BLACK, tm_now->tm_mon + 1, tm_now->tm_mday,
						tm_now->tm_year + 1900, tm_now->tm_hour,
						tm_now->tm_min );
					append_to_textbox_color( chat_window, NULL, buf );
				}

				if ( strcasecmp( tmp, ymsg_sess->user )) {
					sprintf( buf, "%s%s%s%s%s",
						YAHOO_STYLE_BOLDON, YAHOO_COLOR_RED, tmp,
						YAHOO_COLOR_BLACK, YAHOO_STYLE_BOLDOFF );
				} else {
					/* this never really happens, we don't recv our comments */
					sprintf( buf, "%s%s%s%s%s",
						YAHOO_STYLE_BOLDON, YAHOO_COLOR_BLUE, tmp,
						YAHOO_COLOR_BLACK, YAHOO_STYLE_BOLDOFF );
				}

				if ( ! strcmp( ymsg9_field( "124"), "1" ))
					strcat( buf, ": " );
				else
					strcat( buf, " " );

				append_to_textbox_color( chat_window, NULL, buf );

				if ( ! strcmp( ymsg9_field( "124"), "1" )) {
					sprintf( buf, "%s%s%s\n",
						YAHOO_COLOR_BLACK, tmp2, ATTR_RESET );
				} else {
					sprintf( buf, "%s%s%s\n",
						YAHOO_COLOR_PURPLE, tmp2, ATTR_RESET );

					/* check for status change */
					if ( ! strncasecmp( tmp2, "is away ", 8 )) {
						chatter_list_status( tmp, pixmap_status_away, "AW" );
					} else if ( ! strncasecmp( tmp2, "is back", 7 )) {
						chatter_list_status( tmp, pixmap_status_here, "" );
					}
				}
				append_to_textbox( chat_window, NULL, buf );

				break;
		case YMSG9_JOIN:
				/* fixme, need to verify what exactly constitutes a join pkt */
				if ( strcmp( ymsg9_field( "61" ), "" )) {
					/* I joined the room */
					gtk_statusbar_pop( GTK_STATUSBAR(chat_status), st_cid );
					sprintf( buf, "Current Room: %s", ymsg9_field( "104" ));
					gtk_statusbar_push( GTK_STATUSBAR(chat_status),
						st_cid, buf );
					strcpy( ymsg_sess->room, ymsg9_field( "104" ));

					sprintf( buf, "You are now in %s%s%s\n",
						YAHOO_COLOR_BLUE, ymsg9_field( "104" ), ATTR_RESET );
					if ( strcmp( ymsg9_field( "105" ), "" )) {
						strcat( buf, YAHOO_COLOR_CYAN );
						strcat( buf, ymsg9_field( "105" ));
						strcat( buf, ATTR_RESET );
						strcat( buf, "\n" );
					}

					strcpy( tmp, ymsg9_field( "109" ));
					src = tmp;
					ptr = tmp2;
					while( *src ) {
						*ptr = *(src++);
						if ( *ptr == ',' ) {
							ptr++;
							*ptr = ' ';
						}
						ptr++;
					}
					*ptr = '\0';

					strcat( buf, "You see here: " );
					strcat( buf, tmp2 );
					strcat( buf, "\n" );
					append_to_textbox( chat_window, NULL, buf );

					chatter_list_populate( ymsg9_field( "109" ));

					/* little debugging code for showing capabilities */
					/*
					strcpy( tmp, ymsg9_field( "113" ));
					strcpy( tmp2, ymsg9_field( "109" ));
					ptr = tmp;
					src = tmp2;
					while( src ) {
						end = strchr( src, ',' );
						if ( end ) {
							*end = '\0';
							end++;
						}
						end = strchr( ptr, ',' );
						if ( end ) {
							*end = '\0';
							end++;
						}

						fprintf( stderr, "%-20.20s = %x\n", src, atoi( ptr ));
						fflush( stderr );
						if ( ! end ) {
							src = NULL;
						} else {
							src = src + strlen( src ) + 1;
							ptr = ptr + strlen( ptr ) + 1;
						}
					}
					*/
				} else if (( ! strcmp( ymsg9_field( "108" ), "1" )) &&
					( strcasecmp( ymsg9_field( "109" ), ymsg_sess->user ))) {
					/* someone else joined the room while I'm in there */
					strcpy( tmp, ymsg9_field( "109" ));
					chatter_list_add( tmp );

					/* keep track of last thing the user said */
					count = set_last_comment( tmp, "Entered Room" );
					if (( ignore_on_mults ) &&
						( count > ignore_on_mults ) &&
						( ! ignore_check( tmp ))) {
						ignore_toggle( tmp );
						sprintf( buf,
							"Gyach - auto-ignored %s%s%s, entered/left "
							"room %d times without commenting\n",
							YAHOO_COLOR_RED, tmp,
							YAHOO_COLOR_BLACK, count );
						append_to_textbox( chat_window, NULL, buf );
					}
					if (( mute_on_mults ) &&
						( count > mute_on_mults ) &&
						( ! mute_check( tmp ))) {
						mute_toggle( tmp );
						sprintf( buf,
								"Gyach - auto-muted %s%s%s, entered/left "
								"room %d times without commenting\n",
								YAHOO_COLOR_RED, tmp,
								YAHOO_COLOR_BLACK, count );
								append_to_textbox( chat_window, NULL, buf );
					}
					/* don't do this for now because we don't check leave msgs
					if (( suppress_mult ) &&
						( count > 1 )) {
						return;
					}
					*/

					if (( ! ignore_check( tmp )) &&
						( ! mute_check( tmp )) &&
						( show_enters )) {
						sprintf( buf, "%s%s %senters%s the room\n",
							YAHOO_COLOR_BLUE, tmp,
							YAHOO_COLOR_GREEN, YAHOO_COLOR_BLACK );
						append_to_textbox( chat_window, NULL, buf );
					}
				}
				break;
		case YMSG9_EXIT:
				if ( strcasecmp( ymsg_sess->user, ymsg9_field( "109"))) {
					/* someone else left */
					strcpy( tmp, ymsg9_field( "109" ));
					chatter_list_remove( tmp );

					if (( ! ignore_check( tmp )) &&
						( ! mute_check( tmp )) &&
						( show_enters )) {
						sprintf( buf, "%s%s %sleaves%s the room\n",
							YAHOO_COLOR_BLUE, tmp,
							YAHOO_COLOR_GREEN, YAHOO_COLOR_BLACK );
						append_to_textbox( chat_window, NULL, buf );
					}
				} else {
					/* packet that shows up when I leave */
				}
				break;

		case YMSG9_PM:
		case YMSG9_PM_RECV:
				strcpy( tmp, ymsg9_field( "4" ));
				strcpy( tmp2, ymsg9_field( "14" ));

				/* don't think we should get any of these, but just in cas */
				if ( ! strcmp( ymsg9_field( "49" ), "TYPING" )) {
					return;
				}

				if ( ignore_check( tmp )) {
					return;
				}

				if ( mute_check( tmp )) {
					return;
				}

				/* check the fields determining who to accept PMs from */
				accept_pm = 0;
				if (( pm_from_all ) ||
					(( pm_from_users ) &&
					 ( find_user_row( tmp ) >= 0 )) ||
					(( pm_from_friends ) &&
					 ( find_friend( tmp )))) {
					accept_pm = 1;
				}

				ptr = get_last_comment( tmp );
				if (( first_post_is_pm ) &&
					(( ptr == NULL ) ||
					 ( !strcmp( ptr, "Entered Room" )))) {
					if ( ! ignore_check( tmp )) {
						ignore_toggle( tmp );
					}
					sprintf( buf, "Gyach - auto-ignored %s%s%s, first "
						"post was PM: '%s'\n",
						YAHOO_COLOR_RED, tmp, YAHOO_COLOR_BLACK, tmp2 );
					append_to_textbox( chat_window, NULL, buf );
					return;
				}

				if (( url_from_nonroom_user ) &&
					(( strstr( tmp2, "http://" )) ||
					 ( strstr( tmp2, "ftp://" )) ||
					 ( strstr( tmp2, " www." ))) &&
					( find_user_row( tmp )) < 0 ) {
					if ( ! ignore_check( tmp )) {
						ignore_toggle( tmp );
					}
					sprintf( buf, "Gyach - auto-ignored %s%s%s, URL in "
						"PM from user not in room: '%s'\n",
						YAHOO_COLOR_RED, tmp, YAHOO_COLOR_BLACK, tmp2 );
					append_to_textbox( chat_window, NULL, buf );
					return;
				}

				if ( ! accept_pm ) {
					if ( find_friend( tmp )) {
						sprintf( buf, "Ignored PM from Friend (%s%s%s)\n",
							YAHOO_COLOR_RED, tmp, YAHOO_COLOR_BLACK );
						append_to_textbox( chat_window, NULL, buf );
					}
					/* possibly enable this code sometime.
					else if ( find_user_row( tmp ) >= 0 ) {
						sprintf( buf, "Ignored PM from User in Room (%s%s%s)\n",
							YAHOO_COLOR_RED, tmp, YAHOO_COLOR_BLACK );
						append_to_textbox( chat_window, NULL, buf );
					}
					*/
					return;
				}

				if (( ptr = auto_ignore_check( tmp2 )) != NULL ) {
					if ( ! ignore_check( tmp )) {
						ignore_toggle( tmp );
					} else {
						return;
					}

					sprintf( buf, "Gyach - auto-ignored %s%s%s, regex '%s'\n",
						YAHOO_COLOR_RED, tmp, YAHOO_COLOR_BLACK, ptr );
					append_to_textbox( chat_window, NULL, buf );

					if ( ! disp_auto_ignored ) {
						return;
					}
				} 

				if (( ! pm_in_sep_windows ) ||
					( ignore_check( tmp ))) {
					if ( chat_timestamp ) {
						time_now = time( NULL );
						tm_now = localtime( &time_now );
						sprintf( buf, "%s[%d/%d/%d %02d:%02d] ",
							YAHOO_COLOR_BLACK,
							tm_now->tm_mon + 1, tm_now->tm_mday,
							tm_now->tm_year + 1900,
							tm_now->tm_hour, tm_now->tm_min );
						append_to_textbox_color( chat_window, NULL, buf );
					}

					sprintf( buf, "%s%s%s%s%s <private to %s%s%s%s%s> %s%s%s\n",
						YAHOO_COLOR_RED, YAHOO_STYLE_BOLDON, tmp,
						YAHOO_COLOR_BLACK, YAHOO_STYLE_BOLDOFF,
						YAHOO_COLOR_BLUE, YAHOO_STYLE_BOLDON, ymsg_sess->user,
						YAHOO_COLOR_BLACK, YAHOO_STYLE_BOLDOFF,
						YAHOO_STYLE_BOLDON, tmp2, YAHOO_STYLE_BOLDOFF );

					append_to_textbox_color( chat_window, NULL, buf );
				} else {
					/* here we would send to a private pm session */
					/* duplicate to both for now for testing */
					if (( pm_lptr = find_pm_session( tmp )) != NULL ) {
						pm_sess = pm_lptr->data;
					} else {
						/* open a new window if one doesn't exist already */
						pm_sess = (PM_SESSION *)malloc( sizeof( PM_SESSION ));
						pm_sess->pm_window = create_pm_session();
						pm_sess->pm_text = lookup_widget( pm_sess->pm_window,
							"pms_text" );
						pm_sess->pm_user = strdup( tmp );
						pm_sess->typing = 0;

						sprintf( buf, "* %s - Gyach", tmp );
						gtk_window_set_title( GTK_WINDOW(pm_sess->pm_window), buf );

						pm_list = g_list_append( pm_list, pm_sess );

						/* setup the callback for url highlighting */
						gtk_signal_connect_after (GTK_OBJECT (pm_sess->pm_text),
							"motion_notify_event",
							GTK_SIGNAL_FUNC(chat_motion_notify), NULL);

						GTK_WIDGET_UNSET_FLAGS( GTK_WIDGET(pm_sess->pm_text),
							GTK_CAN_FOCUS);

						gtk_widget_set_events( GTK_WIDGET(pm_sess->pm_text),
							GDK_POINTER_MOTION_MASK
							| GDK_POINTER_MOTION_HINT_MASK );
					}
					gtk_widget_show( pm_sess->pm_window );

					if ( auto_raise_pm ) {
						gdk_window_raise( pm_sess->pm_window->window );
					}

					if ( chat_timestamp_pm ) {
						time_now = time( NULL );
						tm_now = localtime( &time_now );
						sprintf( buf, "%s[%d/%d/%d %02d:%02d] ",
							YAHOO_COLOR_BLACK,
							tm_now->tm_mon + 1, tm_now->tm_mday,
							tm_now->tm_year + 1900,
							tm_now->tm_hour, tm_now->tm_min );
						append_to_textbox_color( pm_sess->pm_window,
							pm_sess->pm_text, buf );
					}

					sprintf( buf, "%s%s%s%s%s: %s\n",
						YAHOO_COLOR_RED, YAHOO_STYLE_BOLDON, tmp,
						YAHOO_COLOR_BLACK, YAHOO_STYLE_BOLDOFF, tmp2 );
					append_to_textbox_color( pm_sess->pm_window,
						pm_sess->pm_text, buf );
					sprintf( buf, "* %s - Gyach", tmp );
					gtk_window_set_title( GTK_WINDOW(pm_sess->pm_window), buf );
				}

				/* auto-reply if necessary */
				if (( my_status ) &&
					( auto_reply_when_away ) &&
					( ! ignore_check( tmp )) &&
					( auto_reply_msg )) {
					ymsg9_pm( ymsg_sess, tmp, auto_reply_msg );
				}
				break;

		case YMSG9_NOTIFY:
				strcpy( tmp, ymsg9_field( "4" ));

				/* message that tell us the user is typing to us or stopped */
				/* typing is stored in '49' field */
				if ( ! strcmp( ymsg9_field( "49" ), "TYPING" )) {
					/* this code broken right now, we don't seem to receive */
					/* the typing start messages */
					if (( pm_lptr = find_pm_session( tmp )) != NULL ) {
						pm_sess = pm_lptr->data;
						if ( GTK_WINDOW(pm_sess->pm_window)->title[0] == '*' ) {
							strcpy( buf, "* " );
						} else {
							buf[0] = '\0';
						}
						strcat( buf, tmp );
						if ( ! strcmp( ymsg9_field( "13" ), "1" )) {
							strcat( buf, " [TYPING]" );
						}
						strcat( buf, " - Gyach" );
						gtk_window_set_title( GTK_WINDOW(pm_sess->pm_window),
							buf );
					}
					return;
				}
				break;

		case YMSG9_BUDDY_ON:
				if ( ! strcmp( ymsg9_field( "7" ), "" )) {
					return;
				}

				if ( strcmp( ymsg9_field( "8" ), "" )) {
					/* initial buddy list of people online */
					strcpy( buf, ymsg9_field( "7"));
					ptr = buf;
					last = 0;
					while( ptr ) {
						end = strchr( ptr, ',' );
						if ( end ) {
							*end = '\0';
						} else {
							last = 1;
						}

						add_online_friend( ptr );

						if ( last ) {
							ptr = NULL;
						} else {
							ptr = end + 1;
						}
					}

					show_friends();
					return;
				}
			
				strcpy( tmp, ymsg9_field( "7" ));

				if ( ! strcasecmp( tmp, ymsg_sess->user )) {
					return;
				}

				sprintf( buf, "%sBuddy:%s %s%s%s has logged ON.\n",
					YAHOO_COLOR_BLUE, YAHOO_COLOR_BLACK, YAHOO_COLOR_RED,
					tmp, YAHOO_COLOR_BLACK );

				display = 1;
				if ( find_online_friend( tmp )) {
					display = 0;
				} else {
					add_online_friend( tmp );
				}

				if ( display ) {
					append_to_textbox_color( chat_window, NULL, buf );
					if (( pm_lptr = find_pm_session( tmp )) != NULL ) {
						pm_sess = pm_lptr->data;
						append_to_textbox_color( pm_sess->pm_window,
							pm_sess->pm_text, buf );
					}
				}
				break;

		case YMSG9_BUDDY_OFF:
				if ( ! ymsg_sess->pkt.size ) {
					ymsg_sess->pkt.type = YMSG9_LOGOUT;
					show_yahoo_packet();
				}
				if ( ! strcmp( ymsg9_field( "7" ), "" )) {
					return;
				}
				strcpy( tmp, ymsg9_field( "7"));

				if ( ! strcasecmp( tmp, ymsg_sess->user )) {
					return;
				}

				sprintf( buf, "%sBuddy:%s %s%s%s has logged OFF.\n",
					YAHOO_COLOR_BLUE, YAHOO_COLOR_BLACK, YAHOO_COLOR_RED,
					tmp, YAHOO_COLOR_BLACK );

				display = 1;
				if ( find_online_friend( tmp )) {
					remove_online_friend( tmp );
				} else {
					display = 0;
				}

				if ( display ) {
					append_to_textbox_color( chat_window, NULL, buf );
					if (( pm_lptr = find_pm_session( tmp )) != NULL ) {
						pm_sess = pm_lptr->data;
						append_to_textbox_color( pm_sess->pm_window,
							pm_sess->pm_text, buf );
					}
				}
				break;

		case YMSG9_INVITE: 
				strcpy( tmp, ymsg9_field( "119" ));
				strcpy( tmp2, ymsg9_field( "104" ));

				if ( ! strcmp( tmp, "" )) {
					sprintf( buf, "%sUser has been invited to %s%s\n",
						YAHOO_COLOR_PURPLE, ymsg_sess->room,
						YAHOO_COLOR_BLACK );
					append_to_textbox_color( chat_window, NULL, buf );
					return;
				}
			
				sprintf( buf,
					"%sYou have been invited to (%s%s%s%s%s) by %s%s%s%s%s\n",
					YAHOO_COLOR_PURPLE, YAHOO_COLOR_RED, YAHOO_STYLE_BOLDON,
					tmp2, YAHOO_STYLE_BOLDOFF, YAHOO_COLOR_PURPLE,
					YAHOO_COLOR_BLUE, YAHOO_STYLE_BOLDON, tmp,
					YAHOO_STYLE_BOLDOFF, YAHOO_COLOR_BLACK );
				if (( ignore_check( tmp )) ||
					( disp_auto_ignored ))
					append_to_textbox_color( chat_window, NULL, buf );
				break;

		case YMSG9_FILETRANSFER:
				strcpy( tmp, ymsg9_field( "4" ));

				if ( ! strcmp( tmp, "FILE_TRANSFER_SYSTEM" )) {
					/* we sent a file */
					strcpy( tmp2, ymsg9_field( "14" ));
					sprintf( buf, "%sFILE:%s %s\n",
						YAHOO_COLOR_GREEN, YAHOO_COLOR_BLACK, tmp2 );
					append_to_textbox_color( chat_window, NULL, buf );
				} else {
					/* we received a file */
					strcpy( tmp2, ymsg9_field( "20" ));
					strcpy( tmp3, ymsg9_field( "14" ));
					exp_time = atoi( ymsg9_field( "38" ));
					sprintf( buf, "%sFILE:%s %s%s has sent you a file with "
						"the message: '%s'.\n"
						"        %s\n"
						"        It is available until: %s",
						YAHOO_COLOR_GREEN, YAHOO_COLOR_BLUE, tmp, 
						YAHOO_COLOR_BLACK, tmp3, tmp2, ctime( &exp_time ));
					append_to_textbox_color( chat_window, NULL, buf );
				}
				break;

		case YMSG9_LOGOUT:
				close( ymsg_sess->sock );
				ymsg_sess->sock = -1;
				logged_in = 0;
				set_menu_connected( 0 );

				if ( ymsg_sess->io_callback_tag != 0) {
					gdk_input_remove( ymsg_sess->io_callback_tag );
					ymsg_sess->io_callback_tag = 0;
				}

				if ( ymsg_sess->ping_callback_tag != 0) {
					gdk_input_remove( ymsg_sess->ping_callback_tag );
					ymsg_sess->ping_callback_tag = 0;
				}

				time_now = time(NULL);

				hours = (int)(( time_now - connect_time ) / 3600);
				minutes = (int)(( time_now - connect_time ) % 3600 ) / 60;
				seconds = (int)(( time_now - connect_time ) % 60 );
				connect_time = 0;

				/* show user that we aren't logged in anymore */
				sprintf( buf, "%s%s%s%s%s  Online for %d:%02d:%02d.\n\n",
					YAHOO_COLOR_BLUE, YAHOO_STYLE_BOLDON,
					"Disconnected from Chat!",
					YAHOO_STYLE_BOLDOFF, YAHOO_COLOR_BLACK,
					hours, minutes, seconds );
				append_to_textbox_color( chat_window, NULL, buf );

				gtk_statusbar_pop( GTK_STATUSBAR(chat_status), st_cid );
				gtk_statusbar_push( GTK_STATUSBAR(chat_status), st_cid,
					"Not currently connected to chat..." );

				if ( auto_reconnect ) {
					login_to_yahoo_chat();
				} 
				break;
	}

}

int main( int argc, char **argv ) {
	char gtkrc_filename[1024];
	struct stat gtkrc_stat;
#ifdef OS_WINDOWS
	struct timeval tv;
	fd_set set;
	int ret;
#endif

	DBG( 1, "Gyach v%s\n", VERSION );
	DBG( 1, "%s\n", GYACH_URL );
	DBG( 1, "Debug Level set to: %d\n", DBG_LEVEL );

	memset( ymsg_sess ,0 ,sizeof(YMSG9_SESSION)); 
	ymsg_sess->sock = -1; 

	gtk_set_locale ();

	gtk_init (0,0);

#ifdef OS_WINDOWS
	sprintf( gtkrc_filename, "./gtkrc" );
#else
	sprintf( gtkrc_filename, "%s/.gyach/gtkrc", getenv( "HOME" ));
#endif
	if ( ! stat( gtkrc_filename, &gtkrc_stat )) {
		gtk_rc_parse( gtkrc_filename );
	}

	gyach_init();

	chat_window = build_chat_window();
	gtk_widget_show( chat_window );

	login_window = build_login_window();

	if ( ! auto_login ) {
		gtk_widget_show( login_window );
	} else {
		login_to_yahoo_chat();
	}

	gdk_threads_enter();

#ifdef OS_WINDOWS
	/* since we can't use an io callback with win32, we have to do a */
	/* main loop ourselves and check the socket with a select().     */
	while( ! ymsg_sess->quit ) {
		while( gtk_events_pending()) {
			gtk_main_iteration();
		}

		/* now check for input on our socket */
		FD_ZERO( &set );
		if ( ymsg_sess->sock != -1 )
			FD_SET( ymsg_sess->sock, &set );

		tv.tv_sec = 0;
		tv.tv_usec = 50000;
		ret = select( ymsg_sess->sock + 1, &set, NULL, NULL, &tv );

		if ( ret ) {
			if ( ymsg9_recv_data( ymsg_sess )) {
				show_yahoo_packet();
			}
		}
	}
#else
	gtk_main();
#endif

	return( 0 );
}

